//===================================RISCKER====================================
//=====================================FSM======================================
//------------------------------------------------------------------------------
//File Name:	FSM_Control.v
//Purpose:		HCU Finite State Machine
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------
module FSM_Control(
	input fsm_clk,
	input [3:0]opcode,
	input [7:0]funct,
	input zero,sign, reset,
	input interruption,retfi,
	//Combinational Control Signals
	output [1:0]pc_src, output r2_src,alu_src, output[10:0]operation, output [1:0]reg_src,
	//Sequential Control Signals
	output ipc_write,intoggle,pc_next,ir_write,rf_write,alu_reg,mem_write
	);


//===Opcode parametrization===

//R-Type
parameter 	R_Type =4'b0000;		
//I-Type Operations
parameter	addi = 	4'b1000;
parameter	andi = 	4'b1001;
parameter	ori = 	4'b1010;
parameter	xori = 	4'b1011;
parameter	nandi =	4'b1100;
parameter	nori =	4'b1101;
parameter	multi =	4'b1110;
parameter	divi =	4'b1111;
//I-Type
parameter	j = 	4'b0001;
parameter	sw =	4'b0010;
parameter	lw = 	4'b0011;
parameter	blt = 	4'b0100;
parameter	beq = 	4'b0101;
parameter	jrlw = 	4'b0110;
parameter	jal = 	4'b0111;

//==========================Finite State Machine States Declaration=============================

//							ipc_write intoggle pc_next ir_write rf_write alu_reg mem_write
parameter IDLE = 			7'b0________0_______0________0_______0______0________0;
parameter FETCH = 			7'b0________0_______0________1_______0______0________0;
parameter EXECUTE_OP= 		7'b0________0_______1________0_______0______1________0;
parameter EXECUTE_JRLW= 	7'b0________0_______1________0_______1______0________0;
parameter WRITE_OP= 		7'b0________0_______0________0_______1______0________0;
parameter WRITE_SW= 		7'b0________0_______0________0_______0______0________1;
parameter JUMP= 			7'b0________0_______1________0_______0______0________0;
parameter INTERRUPT=		7'b1________1_______0________0_______0______0________0;
parameter INTJUMP=			7'b1________0_______1________0_______0______0________0;
parameter RETFI=			7'b0________1_______0________0_______0______0________0;

reg [6:0] state = IDLE;

//========================Cycle Control Signals Assignment======================
assign {ipc_write,intoggle,pc_next,ir_write,rf_write,alu_reg,mem_write} = state;

//=======================Finite State Machine===============================
always@(posedge fsm_clk or negedge reset)
if (reset==0)
	state <= IDLE;
else
	  
case (state)
 
IDLE:					state <= FETCH;
 
FETCH:				case (opcode)
						
						R_Type,addi,andi,
						ori,xori,nandi,nori,
						multi,divi,sw,lw:			state <= EXECUTE_OP;
						
						blt,beq,j:				state <= JUMP;
						
						jal:						state <= WRITE_OP;
						
						jrlw:						state <= EXECUTE_JRLW;
						
						endcase
 
EXECUTE_OP:			if (opcode==sw)
							state <= WRITE_SW;
						else
							state <= WRITE_OP;
						
												
EXECUTE_JRLW:		if (retfi)
							state <= RETFI;						
						else if(interruption)
							state <= INTERRUPT;
						else
							state <= FETCH;
 
WRITE_OP:			if (opcode==jal)
							state <= JUMP;
						else if(interruption)
							state <= INTERRUPT;
						else
							state <= FETCH;
 
WRITE_SW:			if(interruption)
							state <= INTERRUPT;
						else
							state <= FETCH;
 
JUMP:					if(interruption)
							state <= INTERRUPT;
						else
							state <= FETCH;

INTERRUPT:			state <= INTJUMP;

INTJUMP:				state <= FETCH;

RETFI:				state <= FETCH;

default:			state <= IDLE;

endcase

/////////////////////////////////////////////////////////////////////////////////////////
//================================INSTRUCTION DECODE=====================================
/////////////////////////////////////////////////////////////////////////////////////////
 
//===========Function parametrization===================
wire [2:0]aluop = opcode[3] ?  opcode[2:0] : funct[2:0];
wire [3:0]shamt = funct[3:0];
wire [1:0]shift_mode = funct[5:4];
wire [1:0]sel = funct[7:6]; //ALU,Shifter,Mul,Div

//=====================Branch conditions (Opcode&Condition)========================
//Signal "branch" selects next PC: Target when it's true and PC+1 when false
wire [1:0]branch = ((opcode==blt) & sign) | ((opcode==beq) & zero) ? 2'b01 : 2'b00;

reg [16:0]control_bus;

//=============================Logical Control Bus assignments==========================
always @ *
case (opcode)
						//		pc_src: Choose next PC address source: PC+1, Target, Register Data 1 or Interruption Address
						//		|		r2_src:	Choose between Rs and Rt as 2nd Register read from Register-File
						//		|		|		alu_src: Choose between Immediate and Register as 2nd ALU operand
						//		|		|		|		operation: Choose ALU Operation(and,or,add,sub,etc), Shift Amount,
						//								Shift Mode(LSL,LSR,ASR,ROT), Select Output source (ALU,Shift,MUL,DIV)
						//		|		|		|		^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
	addi,andi,ori,xori,	//		|		|		|		|		|		|			|		reg_src: Register-File destination data source
	nandi,nori:	control_bus = {2'b00,	1'b0,	1'b1,	aluop,	shamt,	shift_mode,	2'b00,	2'b00};
	
	R_Type:		control_bus = {2'b00,	1'b0,	1'b0,	aluop,	shamt,	shift_mode,	sel,	2'b00};
					
	multi:		control_bus = {2'b00,	1'b0,	1'b1,	aluop,	shamt,	shift_mode,	2'b10,	2'b00};
	
	divi:		control_bus = {2'b00,	1'b0,	1'b1,	aluop,	shamt,	shift_mode,	2'b11,	2'b00};
	
	j:		control_bus = {2'b01,	1'b1,	1'b0,	aluop,	shamt,	shift_mode,	sel,	2'b00};
	
	sw,lw:		control_bus = {2'b00,	1'b1,	1'b1,	3'b000,	shamt,	shift_mode,	2'b00,	2'b10};
	
	blt,beq:	control_bus = {branch,	1'b1,	1'b0,	3'b110,	shamt,	shift_mode,	2'b00,	2'b00};
					
	jrlw:		control_bus = {2'b10,	1'b0,	1'b0,	aluop,	shamt,	shift_mode,	sel,	2'b11};
	
	jal:		control_bus = {2'b01,	1'b0,	1'b0,	aluop,	shamt,	shift_mode,	sel,	2'b01};
	
	default:	control_bus = {2'b00,	1'b0,	1'b0,	aluop,	shamt,	shift_mode,	sel,	2'b00};
endcase
assign {r2_src,alu_src,operation,reg_src} = control_bus[14:0]; 

//Next PC (pc_src) is forced to 2'b11(Interruption Service Address) when an interruption is recognized
assign pc_src = ipc_write ? 2'b11 : control_bus[16:15]; 				


endmodule
