//===================================RISCKER====================================
//====================================UART======================================
//------------------------------------------------------------------------------
//File Name:	UART_FullDuplex.v
//Purpose:		FSM for UART full duplex communication of one byte
//Parameters:	None
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------
module UART_FullDuplex (
	input 	clk,
	//Transmitter
	input 	Tx_Start,
	input 	[7:0]Tx_Data,
	output 	Tx,
	output	TxD_Busy,
	//Receiver
	input 	Rx,
	output	[7:0]Rx_Data,
	output	RxD_Ready
	);

//========================Tx/Rx Control=====================

reg		Frame_Rx=0;						//Frame is being received
reg		Frame_Tx=0;						//Frame is being transmitted

assign	TxD_Busy = Frame_Tx;
assign 	RxD_Ready= ~Frame_Rx;

//=====================Baud Rate Generation==================
parameter clk_in_freq = 50000000;
parameter baud_rate = 9600;
localparam counter_limit = clk_in_freq/(2*baud_rate);

//Tx Baud Rate
reg [15:0]Tx_div_clk_counter=0;
reg Tx_clk_count_reset=0;
reg Tx_baud_clk=0;

always @ (posedge clk or posedge Tx_clk_count_reset)
	if (Tx_clk_count_reset)
		Tx_div_clk_counter <= 0;
	else
		begin
			if (Frame_Tx)
				Tx_div_clk_counter <= Tx_div_clk_counter+1;
		end		

always @ (negedge clk)
	Tx_clk_count_reset <= (Tx_div_clk_counter==counter_limit-1);
	
always @ (posedge Tx_clk_count_reset)
	Tx_baud_clk <= ~Tx_baud_clk;
		
//Rx Baud Rate
reg [15:0]Rx_div_clk_counter=0;
reg Rx_clk_count_reset=0;
reg Rx_baud_clk=0;

always @ (posedge clk or posedge Rx_clk_count_reset)
	if (Rx_clk_count_reset)
		Rx_div_clk_counter <= 0;
	else
		begin
			if (Frame_Rx)
				Rx_div_clk_counter <= Rx_div_clk_counter+1;
		end		

always @ (negedge clk)
	Rx_clk_count_reset <= (Rx_div_clk_counter==counter_limit-1);
	
always @ (posedge Rx_clk_count_reset)
	Rx_baud_clk <= ~Rx_baud_clk;
		
//======================Tx========================

//Tx Data Frame Counter

reg [3:0]Tx_counter=0;
	
always @ (negedge Tx_baud_clk or negedge Frame_Tx)
	if (~Frame_Tx)
		Tx_counter <= 0;
	else
		Tx_counter <= Tx_counter+4'd1;

wire Tx_Counter_Reset = (Tx_counter==10);

//Tx Detection

always@(posedge Tx_Start or posedge Tx_Counter_Reset)
	if (Tx_Counter_Reset)
		Frame_Tx <= 0;
	else
		Frame_Tx <= 1;

//Tx Shift Register

reg [9:0]OutBuffer = 10'b11111111_0_1;

always @ (posedge Tx_baud_clk or negedge Frame_Tx)
	if (~Frame_Tx)
		OutBuffer <= {Tx_Data,2'b01};
	else
		OutBuffer <= {1'b1,OutBuffer[9:1]};
				
//======================Rx=========================

//Rx Data Frame Counter

reg [3:0]Rx_counter=0;
	
always @ (negedge Rx_baud_clk or negedge Frame_Rx)
	if (~Frame_Rx)
		Rx_counter <= 0;
	else
		Rx_counter <= Rx_counter+4'd1;

wire Rx_Counter_Reset = (Rx_counter==10);
	
//Rx Detection
always@(negedge  Rx or posedge Rx_Counter_Reset)
	if (Rx_Counter_Reset)
		Frame_Rx <= 0;
	else
		Frame_Rx <= 1;
	
//Rx Shift Register

reg [9:0]InBuffer=0;

always @ (posedge Rx_baud_clk)
	if (Frame_Rx)
		InBuffer <= {Rx,InBuffer[9:1]};

//================Output Assignment==============
		
assign Tx = OutBuffer[0];

assign Rx_Data = InBuffer[8:1];

endmodule 