//===================================RISCKER====================================
//====================================UART======================================
//------------------------------------------------------------------------------
//File Name:	UART_top.v
//Purpose:		FSM for UART full duplex communication of all bytes
//Parameters:	TxBytes (# of bytes for transmission) RxBytes (Number of Bytes for reception) baud_rate
//------------------------------------------------------------------------------
//=====================Alfredo Gualdron - Jose P. Pinilla=======================
//=================================Semillero ADT================================
//================================UPB Bucaramanga===============================
//Website:	http://semilleroadt.upbbga.edu.co/xiscker
//E-mail:	semilleroadt@gmail.com
//------------------------------------------------------------------------------
module UART_top #(parameter TxBytes = 26, RxBytes = 3, baud_rate = 9600)
	
	(clk,DataOut,Tx,Rx,DataIn,Trigger,Rx_Last_Byte);
	
	localparam Tx_width = TxBytes*8;
	localparam Rx_width = RxBytes*8;
	localparam Tx_counter_width = $clog2(TxBytes);
	localparam Rx_counter_width = $clog2(RxBytes);
		
	input clk;
	input [Tx_width-1:0]DataOut;
	output Tx;

	input Rx;
	output reg [Rx_width-1:0]DataIn=0;
	input Trigger;
	
//=============Tx================
//Synchronously detect when all bytes have been sent
reg [Tx_counter_width:0]Tx_Byte_Counter=0;
reg Tx_Last_Byte =0;
always @ (posedge clk)
	Tx_Last_Byte <= (Tx_Byte_Counter==TxBytes);

//Detect Trigger and keep transmitter enabled until last byte
reg Enable_Tx=0;
always @ (posedge Trigger or posedge Tx_Last_Byte)
	if (Tx_Last_Byte)
		Enable_Tx <= 0;
	else
		Enable_Tx <= 1;

//Synchronize all Data
reg [Tx_width-1:0] sync_data=0;
always @ (posedge clk)
	sync_data <= DataOut;

//Start Transmition if transmitter isn't busy and transmitter is Enabled
wire TxD_Busy;

reg Tx_Start=0;
always @ (negedge clk)
	if (Enable_Tx)
		Tx_Start <= ~TxD_Busy;
		
//When Transmitter stops being busy increment count
always @ (negedge TxD_Busy or posedge Tx_Last_Byte)
	if(Tx_Last_Byte)
		Tx_Byte_Counter <= 0;
	else if (Enable_Tx)
		Tx_Byte_Counter <= Tx_Byte_Counter+1'b1;

//Data is transmitted by Bytes through a Shift Register while transmitter is enabled
reg [Tx_width-1:0] Tx_Data=0;
always @ (negedge TxD_Busy or negedge Enable_Tx)
	if (~Enable_Tx)
		Tx_Data <= sync_data;
	else
		Tx_Data  <= Tx_Data>>8;

//=============Rx================
//Synchronously detect when all bytes have been received
reg [Rx_counter_width-1:0] Rx_Byte_Counter=0;
output reg Rx_Last_Byte =0;
always @ (negedge clk)
	Rx_Last_Byte <= (Rx_Byte_Counter==RxBytes);

//Fill Rx_Stream Shift Register whenever Rx_Data is Ready
wire RxD_Ready;
wire [7:0] Rx_Data;
reg [Rx_width-1:0] Rx_Stream=0;
always @ (posedge RxD_Ready)
	Rx_Stream <= {Rx_Data,Rx_Stream[Rx_width-1:8]};

//Enable Byte reception count while number of Bytes are received
reg Enable_Rx=0;
always @ (negedge Rx or posedge Rx_Last_Byte)
	if (Rx_Last_Byte)
		Enable_Rx <= 0;
	else
		Enable_Rx <= 1;
		
//Receiver Counter
always @ (posedge RxD_Ready or posedge Rx_Last_Byte)
	if (Rx_Last_Byte)
		Rx_Byte_Counter <= 0;
	else if (Enable_Rx)
		Rx_Byte_Counter <= Rx_Byte_Counter+1'b1;

//Register received stream to avoid changes while being read
always @ (posedge Rx_Last_Byte)
	DataIn <= Rx_Stream;

	
//UART Full-Duplex Transceiver
UART_FullDuplex UART_Byte_Transceiver(
	.clk(clk),
	.Tx_Start(Tx_Start),
	.Tx_Data(Tx_Data[7:0]),
	.Tx(Tx),
	.TxD_Busy(TxD_Busy),
	.Rx(Rx),
	.Rx_Data(Rx_Data),
	.RxD_Ready(RxD_Ready)
	);
	
defparam UART_Byte_Transceiver.baud_rate = baud_rate;

endmodule 